﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using STM.Common.DataAccess.Entities;
using STM.Common.DataAccess;
using STM.Common;
using System.Xml;
using System.Xml.Schema;
using System.IO;
using System.Xml.Linq;
using System.Configuration;
using STM.Common.Exceptions;
using STM.Common.Services.Internal.Interfaces;
using System.Web;
using STM.SSC.Internal.Models;

namespace STM.Common.Services.Internal
{
    public class PublishedRtzMessageService : InternalServiceBase<PublishedRtzMessage, StmDbContext>, IPublishedRtzMessageService
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private string m_path = Path.Combine(Environment.CurrentDirectory, @"..\..\Schema");
        private IACLObjectService _aclService;
        private IVisSubscriptionService _subscriberService;
        private ILogEventService _logEventService;
        private INotificationService _notificationService;

        public PublishedRtzMessageService(StmDbContext dbContext, 
            LogDbContext logContext) : base(dbContext)
        {
            _aclService = new ACLObjectService(dbContext);
            _subscriberService = new VisSubscriptionService(dbContext);
            _logEventService = new LogEventService(logContext);
            _notificationService = new NotificationService(dbContext, logContext);
        }

        public override void Insert(PublishedRtzMessage entity)
        {
            SaveAndUpdate(entity, true);
        }

        public override void Update(PublishedRtzMessage entityToUpdate)
        {
            SaveAndUpdate(entityToUpdate, false);
        }

        private void SaveAndUpdate(PublishedRtzMessage entity, bool isNew)
        {
            try
            {
                Validate(entity);
                Populate(entity);

                if (isNew)
                {
                    base.Insert(entity);
                }
                else
                {
                    base.Update(entity);
                }

            }
            catch (ArgumentOutOfRangeException ex)
            {
                throw ex;
            }
            catch (ArgumentNullException ex)
            {
                throw ex;
            }
            catch (StmSchemaValidationException valEx)
            {
                throw valEx;
            }
            catch (Exception ex)
            {
                throw ex;
            }
        }

        private void Populate(PublishedRtzMessage entity)
        {
            var msg = Serialization.ByteArrayToString(entity.Message);
            var routeStatus = XMLParser.GetRouteStatus(msg);

            if (string.IsNullOrEmpty(routeStatus))
            {
                entity.MessageStatus = RouteStatus.Unknown;
            }
            else if (entity.MessageStatus == RouteStatus.Inactive)
            {
                // If route status 8 in entity do not try to read from xml
            }
            else
            {
                int statusId;
                if (int.TryParse(routeStatus, out statusId))
                {
                    entity.MessageStatus = (RouteStatus)statusId;
                }
            }

            DateTime? validityPeriodStart = XMLParser.GetValidityPeriodStart(msg);
            if (validityPeriodStart == DateTime.MinValue)
            {
                entity.MessageValidFrom = null;
            }
            else
            {
                entity.MessageValidFrom = validityPeriodStart.Value;
            }
            DateTime? validityPeriodStop = XMLParser.GetValidityPeriodStop(msg);
            if (validityPeriodStop == DateTime.MinValue)
            {
                entity.MessageValidTo = null;
            }
            else
            {
                entity.MessageValidTo = validityPeriodStop.Value;
            }
        }

        private void Validate(PublishedRtzMessage entity)
        {
            var msg = Serialization.ByteArrayToString(entity.Message);
            
            var validator = new StmSchemaValidator();

            //validate the XML against its XSD
            validator.ValidateRTZMessageXML(msg);

            //validate the message content according to our business rules
            //var parsedMessage = Serialization.Deserialize<Route>(msg);
            validator.ValidateRTZMessage(msg);

            //Validate UVID between header and RTZ
            validator.ValidateUVID(entity.MessageID, msg);
        }

        public void SendMessageToSubsribers(string message, string dataId)
        {
            //call proxy class to SSC -> callservice for each subscriber 
            var method = "SendMessageToSubsribers";
            var subscriptions = _subscriberService.Get(x => 
                x.MessageID == dataId && x.IsAuthorized, includeProperties: "SubscriberIdentity");

            if (subscriptions == null)
                return;

            foreach (var subscription in subscriptions)
            {
                SendMessage(message, dataId, subscription.CallbackEndpoint, subscription.SubscriberIdentity);
            }
        }

        public void SendMessage(string message, string dataId, string endpoint, Identity identity)
        {
            var client = new SSC.Internal.SccPrivateService();

            try
            {

                var result = client.CallService(new CallServiceRequestObj
                {
                    Body = message,
                    EndpointMethod = endpoint + "?uvid=" + dataId,
                    Headers = new List<Header>
                    {
                        new Header("content-type", "text/xml; charset=UTF8; encoding='utf-8'")
                    },
                    RequestType = "POST",
                });

                if (result.StatusCode != 200)
                {
                    throw new Exception(result.Body);
                }

                string msg = string.Format("Voyageplan with uvid {0} sent to subscriber {1}", dataId, identity.Name);
                log.Info(msg);
            }
            catch (Exception ex)
            {
                log.Error(ex);

                // Send notification
                var notification = new Interfaces.Notification();
                notification.FromName = identity.Name;
                notification.FromId = identity.UID;
                notification.DataId = dataId;
                notification.NotificationType = EnumNotificationType.ERROR_MESSAGE;
                notification.Subject = "Unable to send message to subscriber";
                notification.Body = string.Format("Voyageplan with UVID {0} could not be sent to subscriber {1}, {2}", dataId, identity.Name, identity.UID);
                notification.NotificationSource = EnumNotificationSource.VIS;
                _notificationService.Notify(notification);
            }
        }
    }
}