﻿using STM.Common.DataAccess.Entities;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using STM.Common.DataAccess;
using STM.Common;
using STM.Common.Services.Internal.Interfaces;
using Newtonsoft.Json;
using STM.SSC.Internal.Models;

namespace STM.Common.Services.Internal
{
    public class UploadedMessageService : InternalServiceBase<UploadedMessage, StmDbContext>, IUploadedMessageService
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
        private INotificationService _notificationService;

        public UploadedMessageService(StmDbContext dbContext,
            INotificationService notificationService) : base(dbContext)
        {
            _notificationService = notificationService;
        }

        public void InsertRTZ(UploadedMessage entity)
        {
            var msg = Serialization.ByteArrayToString(entity.Message);
            var validator = new StmSchemaValidator();

            //validate the XML against its XSD
            validator.ValidateRTZMessageXML(msg);

            //var parsedMessage = Serialization.Deserialize<Route>(msg);
            //validate the message content according to our business rules
            validator.ValidateRTZMessage(msg);

            //Validate UVID between header and RTZ
            validator.ValidateUVID(entity.MessageID, msg);

            base.Insert(entity);
        }

        public void InsertTXT(UploadedMessage entity)
        {
            var msg = Serialization.ByteArrayToString(entity.Message);
            var validator = new StmSchemaValidator();
            validator.ValidateTextMessageXML(msg);
            //var parsedMessage = Serialization.Deserialize<TextMessage>(msg);

            base.Insert(entity);
        }

        public void InsertArea(UploadedMessage entity)
        {
            var msg = Serialization.ByteArrayToString(entity.Message);
            var validator = new StmSchemaValidator();
            validator.ValidateAreaMessageXML(msg);
            //var parsedMessage = Serialization.Deserialize<Area>(msg);

            base.Insert(entity);
        }

        public void InsertPCM(UploadedMessage message)
        {
            var msg = Serialization.ByteArrayToString(message.Message);
            var validator = new StmSchemaValidator();
            validator.ValidatePCMMessageXML(msg);

            base.Insert(message);
        }

        public List<UploadedMessage> GetMessagesByLimitNumber(int numberOfMessages, List<MessageType> mTypes)
        {
            List<UploadedMessage> result = new List<UploadedMessage>();
            var noMessagesLeft = numberOfMessages;
            foreach(var mType in mTypes)
            {
                var messages = base._context.UploadedMessage.Include("From").Include("MessageType").OrderByDescending(
                    x => x.ReceiveTime).Where(x => x.Notified && !x.FetchedByShip && x.MessageType.ID == mType.ID).
                    Take(noMessagesLeft).ToList();
                noMessagesLeft -= messages.Count();
                foreach(var msg in messages)
                {
                    result.Add(msg);
                }
            }
            return result;
        }

        public List<UploadedMessage> GetAllUnFetchedMessages(List<MessageType> mTypes)
        {
            List<UploadedMessage> result = new List<UploadedMessage>();

            foreach(var mType in mTypes)
            {
                var messages = base._context.UploadedMessage.Include("From").Include("MessageType").
                    OrderByDescending(x => x.ReceiveTime).Where(x => x.Notified && 
                    !x.FetchedByShip && 
                    x.MessageType.ID == mType.ID).ToList();
                foreach (var msg in messages)
                {
                    result.Add(msg);
                }
            }
            return result;
        }

        public List<UploadedMessage> GetUnFetchedMessagesById(string Id, List<MessageType> mTypes)
        {
            List<UploadedMessage> result = new List<UploadedMessage>();

            foreach (var mType in mTypes)
            {
                var messages = base._context.UploadedMessage.Include("From").Include("MessageType").
                    OrderByDescending(x => x.ReceiveTime).Where(x => x.Notified && 
                    !x.FetchedByShip && 
                    x.MessageID == Id && 
                    x.MessageType.ID == mType.ID).ToList();
                foreach(var msg in messages)
                {
                    result.Add(msg);
                }
            }
            return result;
        }

        public List<UploadedMessage> GetMessagesByLimitNumberAndDataId(int numberOfMessages, string Id, List<MessageType> mTypes)
        {
            List<UploadedMessage> result = new List<UploadedMessage>();
            var noMessagesLeft = numberOfMessages;

            foreach (var mType in mTypes)
            {
                var messages = base._context.UploadedMessage.Include("From").Include("MessageType").
                    OrderByDescending(x => x.ReceiveTime).Where(x => x.Notified && 
                    !x.FetchedByShip && 
                    x.MessageID == Id && 
                    x.MessageType.ID == mType.ID).Take(noMessagesLeft).ToList();
                noMessagesLeft -= messages.Count();
                foreach (var msg in messages)
                {
                    result.Add(msg);
                }
            }
            return result;
        }

        public int GetNumberOfRemainingMessages(List<MessageType> mTypes)
        {
            int result = 0;

            foreach (var mType in mTypes)
            {
                var messages = base.Get(x => x.Notified && 
                !x.FetchedByShip && 
                x.MessageType.ID == mType.ID).ToList();
                result += messages.Count();
            }

            return result;
        }

        public int GetNumberOfRemainingMessages(string dataId, List<MessageType> mTypes)
        {
            int result = 0;

            foreach (var mType in mTypes)
            {
                var messages = base.Get(x => x.Notified && 
                !x.FetchedByShip && 
                x.MessageID == dataId &&
                x.MessageType.ID == mType.ID).ToList();
                result += messages.Count();
            }

            return result;
        }

        public void SendAck(IList<UploadedMessage> messages)
        {
            var client = new STM.SSC.Internal.SccPrivateService();

            foreach (var message in messages)
            {
                try
                {
                    if (message.DeliveryAckReqested && !message.AckDelivered)
                    {
                        //var ackObject = new DeliveryAck(message.ID.ToString(), message.MessageID, DateTime.UtcNow, message.From.UID, message.From.Name, InstanceContext.ServiceId, InstanceContext.ServiceName, "Message delivered");
                        var ackObject = new DeliveryAck
                        {
                            AckResult = "OK",
                            FromId = message.From.UID,
                            FromName = message.From.Name,
                            Id =  Guid.NewGuid().ToString(),
                            ReferenceId = message.ID.ToString(),
                            TimeOfDelivery = DateTime.UtcNow,
                            ToId = InstanceContext.ServiceId,
                            ToName = InstanceContext.ServiceName
                        };
                        client.CallService(new CallServiceRequestObj
                        {
                            Body = JsonConvert.SerializeObject(ackObject),
                            EndpointMethod = message.DeliveryAckEndpoint,
                            Headers = new List<Header>
                            {
                                new Header("content-type", "application/json; charset=UTF8;charset=UTF8")
                            },
                            RequestType = "POST",
                        });
                    }
                }
                catch (Exception ex)
                {
                    log.Error(ex);

                    // Send notification
                    var notification = new Internal.Interfaces.Notification();
                    notification.FromName = message.From.Name;
                    notification.FromId = message.From.UID;
                    notification.DataId = message.MessageID;
                    notification.NotificationType = EnumNotificationType.ERROR_MESSAGE;
                    notification.Subject = "Unable to send acknowledgement";
                    notification.Body = string.Format("Acknowledgement for voyageplan with UVID {0} could not be sent to {1}, {2}", message.MessageID, message.From.Name, message.From.UID);
                    notification.NotificationSource = EnumNotificationSource.VIS;
                    _notificationService.Notify(notification);
                }
            }
        }
    }
}
