﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using STM.Common;
using STM.Common.Exceptions;
using System.Reflection;

namespace STM.Common
{
    public class StmSchemaValidator
    {
        private string cachekey = "StmIncomingSchema";
        private static object lockobject = new object();
        private string valResult;

        private SchemaValidator validator { get; set; }

        public StmSchemaValidator()
        {
            lock (lockobject)
            {
                validator = SchemaValidator.GetSchemaValidator(cachekey);
                if (validator == null)
                {
                    var schemaAssembly = Assembly.GetAssembly(typeof(STM.Common.Schema.Schema));
                    var names = schemaAssembly.GetManifestResourceNames();

                    validator = new SchemaValidator(schemaAssembly, names.ToList());
                    SchemaValidator.CacheSchemaValidator(validator, cachekey);
                }
            }
        }

        //TODO: Need to be able to seperate according to message type
        //public VisSchemaValidator(MessageType messageType)
        //{
        //    lock (lockobject)
        //    {
        //        validator = SchemaValidator.GetSchemaValidator(cachekey);

        //        if (validator == null)
        //        {
        //            Assembly schemaAssembly = null;
        //            switch (messageType.Name.ToLower())
        //            {
        //                case "rtz":
        //                    schemaAssembly = Assembly.GetAssembly(typeof(Route));
        //                    break;
        //                case "txtmsg":
        //                    schemaAssembly = Assembly.GetAssembly(typeof(TextMessage));
        //                    break;
        //                case "s124":
        //                    schemaAssembly = Assembly.GetAssembly(typeof(StmMessage));
        //                    break;
        //            }
        //            var names = schemaAssembly.GetManifestResourceNames();

        //            validator = new SchemaValidator(schemaAssembly, names.ToList());
        //            //SchemaValidator.CacheSchemaValidator(validator, cachekey);
        //        }
        //    }
        //}

        public void ValidateAreaMessageXML(string msg)
        {
            valResult = null;
            if (msg == null)
            {
                throw new StmSchemaValidationException("Area message is null. Not a valid message.");
            }

            valResult = validator.ValidateXML(msg);

            if (!string.IsNullOrEmpty(valResult))
            {
                throw new StmSchemaValidationException("Area message does not validate ok. Details: " + valResult);
            }
        }

        /// <summary>
        /// Validate all parts of a RTZ message
        /// </summary>
        /// <param name="msg"></param>
        public void ValidateRTZMessageXML(string msg)
        {
            valResult = null;
            if (msg == null)
            {
                throw new StmSchemaValidationException("Message is null. Not a valid message.");
            }

            valResult = validator.ValidateXML(msg);

            if (!string.IsNullOrEmpty(valResult))
            {
                throw new StmSchemaValidationException("Message does not validate ok. Details: " + valResult);
            }
        }

        public void Validate()
        {

        }


        /// <summary>
        /// Validate all parts of a RTZ message
        /// </summary>
        /// <param name="msg"></param>
        public void ValidateRTZMessage(string msg)
        {
            // Reset
            valResult = "";

            if (msg == null)
            {
                throw new StmSchemaValidationException("STM message is null. Not a valid message.");
            }

            string routeInfo = XMLParser.GetRouteInfo(msg);
            string wayPoints = XMLParser.GetWayPoints(msg);
            string schedules = XMLParser.GetSchedules(msg);
            if (string.IsNullOrEmpty(routeInfo) || 
                string.IsNullOrEmpty(wayPoints) || 
                string.IsNullOrEmpty(schedules))
            {
                throw new StmSchemaValidationException("Some or all parts of the STM message are null. Not a valid message.");
            }

            // TODO: Validate correct version
            //if(msg.version != schemaVersion)

            //TODO: Validate routeinfo content
            string routeStatus = XMLParser.GetRouteStatus(msg);
            if(!string.IsNullOrEmpty(routeStatus))
            {
                int status = Convert.ToInt32(routeStatus);
                if(status < 1 || status > 8)
                {
                    throw new ArgumentOutOfRangeException("routeInfo.routeStatus", status, "Forbidden status value.");
                }

            }
            else
            {
                //throw new STMSchemaValidationException("Mandatory routeStatus is missing.");
                throw new ArgumentNullException("routeInfo.routeStatus", "Mandatory element is null.");
            }

            //TODO: Validate schedule?
            //msg.schedules.schedule[0].calculated.sheduleElement[0].

            //TODO: validate waypoints?

            //TODO: validate extensions?

            if (valResult != "")
            {
                throw new StmSchemaValidationException("Message from external does not validate ok. Details: " + valResult);

            }

        }

        public void ValidateUVID(string messageId, string messageRTZ)
        {
            if (!string.IsNullOrEmpty(messageRTZ))
            {
                string routeInfo = XMLParser.GetRouteInfo(messageRTZ);
                if(!string.IsNullOrEmpty(routeInfo))
                {
                    string vesselVoyage = XMLParser.GetVesselVoyage(messageRTZ);
                    if(!string.IsNullOrEmpty(vesselVoyage))
                    {
                        if(vesselVoyage.ToLower() != messageId.ToLower())
                        {
                            throw new StmSchemaValidationException(string.Format("Inconsistent UVID: UVID in argument = {0} must be equal UVID in RTZ = {1}.",
                                messageId, vesselVoyage));
                        }
                    }
                }
            }
            else
            {
                throw new StmSchemaValidationException("Could not read UVID from RTZ message.");
            }
        }

        /// <summary>
        /// Validate all parts of a Text message
        /// </summary>
        /// <param name="msg"></param>
        public void ValidateTextMessage(string msg)
        {
            if (msg == null)
            {
                throw new StmSchemaValidationException("STM message is null. Not a valid message.");
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="msg"></param>
        public void ValidateTextMessageXML(string msg)
        {
            valResult = null;
            if (msg == null)
            {
                throw new StmSchemaValidationException("Text message is null. Not a valid message.");
            }

            valResult = validator.ValidateXML(msg);

            if (!string.IsNullOrEmpty(valResult))
            {
                throw new StmSchemaValidationException("Text message does not validate ok. Details: " + valResult);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="msg"></param>
        public void ValidatePCMMessageXML(string msg)
        {
            valResult = null;
            if (msg == null)
            {
                throw new StmSchemaValidationException("PCM message is null. Not a valid message.");
            }

            valResult = validator.ValidateXML(msg);

            if (!string.IsNullOrEmpty(valResult))
            {
                throw new StmSchemaValidationException("PCM message does not validate ok. Details: " + valResult);
            }
        }


        /// <summary>
        /// Validate a sub-part of STMMessage (e.g. "MAI" or "OAP"). Add validation errors to valResult.
        /// </summary>
        /// <param name="msgPart"></param>
        /// <param name="msgPartName"></param>
        private void ValidateMessagePart(XmlElement msgPart, string msgPartName)
        {
            if (msgPart != null)
            {
                string res = validator.ValidateXML(msgPart.OuterXml);

                if (res != null)
                {
                    valResult += string.Format("{0}-message: {1}", msgPartName, res);
                }
            }
        }
    }
}