﻿using System.Threading.Tasks;
using STM.Common.DataAccess.Entities;
using STM.Common.Services.Internal;
using STM.Common.DataAccess;
using System.ComponentModel.DataAnnotations;
using STM.VIS.Services.Private.Models;
using System.Collections.Generic;
using System.Net.Http;
using System.Net;
using System.Web.Http;
using Microsoft.Practices.Unity;
using System.Data.Entity;
using System;
using STM.Common;
using System.Configuration;
using STM.Common.Services;
using STM.Common.Services.Internal.Interfaces;
using System.Threading;
using System.Web.Http.Controllers;
using System.Linq;

namespace STM.VIS.Services.Private.Controllers
{
    /// <summary>
    /// 
    /// </summary>
    public class GetMessageController : LoggingControllerBase
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        private IUploadedMessageService _uploadedMessageService;
        private IMessageTypeService _messageTypeService;

        /// <summary>
        /// 
        /// </summary>
        /// <param name="uploadedMessageService"></param>
        /// <param name="messageTypeService"></param>
        [InjectionConstructor]
        public GetMessageController(IUploadedMessageService uploadedMessageService, IMessageTypeService messageTypeService)
        {
            _uploadedMessageService = uploadedMessageService;
            _messageTypeService = messageTypeService;
        }

        /// <summary>
        /// Retrieve received messages from VIS database
        /// </summary>
        /// <param name="limitQuery">Limit the number of messages to be received</param>
        /// <param name="dataId">Can be a unique voyage id or any other id for a datapoint</param>
        /// <response code="200">Success</response>
        /// <response code="default">Unexpected error</response>
        [HttpGet]
        [Route("getMessage")]
        public MessageEnvelope GetMessage(string limitQuery = null,
            string dataId = null)
        {
            log.Info("Incoming request to " + GetCurrentMethod());

            if (dataId != null && !FormatValidation.IsValidUvid(dataId))
            {
                throw CreateHttpResponseException(HttpStatusCode.BadRequest, "Invalid UVID format");
            }

            int limitNumberOfMessages = 0;
            if (int.TryParse(limitQuery, out limitNumberOfMessages))
            {
                log.Info(string.Format("LimitQuery parsed successfully to {0}.", limitNumberOfMessages));
            }

            try
            {
                var result = new MessageEnvelope();
                result.Messages = new List<Message>();
                var uploadedMessages = new List<UploadedMessage>();
                var mTypes = _messageTypeService.Get(m => m.Name != "PCM").ToList();

                if (string.IsNullOrEmpty(dataId))
                {
                    if (limitNumberOfMessages > 0)
                    {
                        uploadedMessages = _uploadedMessageService.GetMessagesByLimitNumber(limitNumberOfMessages, mTypes);
                    }
                    else
                    {
                        uploadedMessages = _uploadedMessageService.GetAllUnFetchedMessages(mTypes);
                    }
                }
                else
                {
                    if (limitNumberOfMessages > 0)
                    {
                        uploadedMessages = _uploadedMessageService.GetMessagesByLimitNumberAndDataId(limitNumberOfMessages, dataId, mTypes);
                    }
                    else
                    {
                        uploadedMessages = _uploadedMessageService.GetUnFetchedMessagesById(dataId, mTypes);
                    }
                }

                foreach (var uploadedMessage in uploadedMessages)
                {
                    uploadedMessage.FetchedByShip = true;
                    uploadedMessage.FetchTime = DateTime.UtcNow;
                    _uploadedMessageService.Update(uploadedMessage);

                    var messageToAdd = new Message();
                    messageToAdd.From = uploadedMessage.From.UID;
                    messageToAdd.MessageType = uploadedMessage.MessageType.Name;
                    messageToAdd.ReceivedAt = uploadedMessage.ReceiveTime;
                    var messageBody = Serialization.ByteArrayToString(uploadedMessage.Message);
                    messageToAdd.StmMessage = new StmMessage(messageBody);
                    messageToAdd.Id = string.IsNullOrEmpty(dataId) ? string.Empty : dataId;
                    messageToAdd.CallbackEndpoint = uploadedMessage.CallbackEndpoint;

                    result.Messages.Add(messageToAdd);
                }
                _context.SaveChanges();

                result.NumberOfMessages = uploadedMessages.Count;
                if (string.IsNullOrEmpty(dataId))
                {
                    result.RemainingNumberOfMessages = _uploadedMessageService.GetNumberOfRemainingMessages(mTypes);
                }
                else
                {
                    result.RemainingNumberOfMessages = _uploadedMessageService.GetNumberOfRemainingMessages(dataId, mTypes);
                }

                _uploadedMessageService.SendAck(uploadedMessages);

                _context.SaveChanges();
                return result;
            }
            catch (HttpResponseException ex)
            {
                log.Error(ex);

                throw ex;
            }
            catch (Exception ex)
            {
                log.Error(ex);

                string msg = "VIS internal server error. " + ex.Message;
                throw CreateHttpResponseException(HttpStatusCode.InternalServerError, msg);
            }
        }
    }
}